#include "eldbus_empd_empc.h"

static int _log_main = -1;
#undef ERR
#define ERR(...) EINA_LOG_DOM_ERR(_log_main, __VA_ARGS__);
int EMPD_EMPC_BACKGROUND_CHANGED_EVENT = 0;

void
empd_empc_set_background_file_call(Eldbus_Proxy *proxy, const char *artist, const char *album, const char *file, const char *key)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "SetBackgroundFile");
   if (!eldbus_message_arguments_append(msg, "ssss", artist, album, file, key))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void 
empd_empc_set_background_data_call(Eldbus_Proxy *proxy, Eina_Value *args)
{
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   Eldbus_Message *msg = eldbus_proxy_method_call_new(proxy, "SetBackgroundData");
   if (!eldbus_message_from_eina_value("sssay", msg, args))
     {
        ERR("Error: Filling message from eina value.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empc_get_background(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empc_Get_Background_Cb cb = data;
   const char *error, *error_msg;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   const char *file = NULL;
   const char *key = NULL;
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, file, key);
        return;
     }
   if (!eldbus_message_arguments_get(msg, "ss", &file, &key))
     {
        Eldbus_Error_Info error_info = {"", ""};
        ERR("Error: Getting arguments from message.");
        cb(proxy, user_data, pending, &error_info, file, key);
        return;
     }
   cb(proxy, user_data, pending, NULL, file, key);
   return;
}

Eldbus_Pending *
empd_empc_get_background_call(Eldbus_Proxy *proxy, Empd_Empc_Get_Background_Cb cb, const void *data)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "GetBackground");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empc_get_background, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

static void
empd_empc_background_changed_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empc_BackgroundChanged_Data *s_data = func_data;
   free(s_data->file);
   free(s_data->key);
   free(s_data);
}

static void
on_empd_empc_background_changed(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empc_BackgroundChanged_Data *s_data = calloc(1, sizeof(Empd_Empc_BackgroundChanged_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "ss", &s_data->file, &s_data->key))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   s_data->file = strdup(s_data->file);
   s_data->key = strdup(s_data->key);
   ecore_event_add(EMPD_EMPC_BACKGROUND_CHANGED_EVENT, s_data, empd_empc_background_changed_data_free, NULL);
}

void
empd_empc_log_domain_set(int id)
{
   _log_main = id;
}

void
empd_empc_proxy_unref(Eldbus_Proxy *proxy)
{
   Eldbus_Object *obj = eldbus_proxy_object_get(proxy);
   eldbus_proxy_unref(proxy);
   eldbus_object_unref(obj);
}

Eldbus_Proxy *
empd_empc_proxy_get(Eldbus_Connection *conn, const char *bus, const char *path)
{
   Eldbus_Object *obj;
   Eldbus_Proxy *proxy;
   EINA_SAFETY_ON_NULL_RETURN_VAL(conn, NULL);
   EINA_SAFETY_ON_NULL_RETURN_VAL(bus, NULL);
   if (!path) path = "/";
   obj = eldbus_object_get(conn, bus, path);
   proxy = eldbus_proxy_get(obj, "org.empd.empc");
   eldbus_proxy_signal_handler_add(proxy, "BackgroundChanged", on_empd_empc_background_changed, proxy);
   if (!EMPD_EMPC_BACKGROUND_CHANGED_EVENT)
     EMPD_EMPC_BACKGROUND_CHANGED_EVENT = ecore_event_type_new();
   return proxy;
}
