#include "eldbus_empd_empdd.h"

static int _log_main = -1;
#undef ERR
#define ERR(...) EINA_LOG_DOM_ERR(_log_main, __VA_ARGS__);
int EMPD_EMPDD_STATUS_EVENT = 0;
int EMPD_EMPDD_CURRENT_SONG_EVENT = 0;
int EMPD_EMPDD_QUEUE_LIST_EVENT = 0;
int EMPD_EMPDD_LOGIN_FAILED_EVENT = 0;
int EMPD_EMPDD_PERMISSION_FAILED_EVENT = 0;
int EMPD_EMPDD_CONNECTED_EVENT = 0;
int EMPD_EMPDD_DISCONNECTED_EVENT = 0;
int EMPD_EMPDD_DATABASE_UPDATE_BEGIN_EVENT = 0;
int EMPD_EMPDD_DATABASE_UPDATE_END_EVENT = 0;

void
empd_empdd_connect_call(Eldbus_Proxy *proxy, const char *host, int port, const char *password)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Connect");
   if (!eldbus_message_arguments_append(msg, "sis", host, port, password))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empdd_is_connected(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_Is_Connected_Cb cb = data;
   const char *error, *error_msg;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   Eina_Bool isconnected = 0;
   const char *host = NULL;
   unsigned int port = 0;
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, isconnected, host, port);
        return;
     }
   if (!eldbus_message_arguments_get(msg, "bsu", &isconnected, &host, &port))
     {
        Eldbus_Error_Info error_info = {"", ""};
        ERR("Error: Getting arguments from message.");
        cb(proxy, user_data, pending, &error_info, isconnected, host, port);
        return;
     }
   cb(proxy, user_data, pending, NULL, isconnected, host, port);
   return;
}

Eldbus_Pending *
empd_empdd_is_connected_call(Eldbus_Proxy *proxy, Empd_Empdd_Is_Connected_Cb cb, const void *data)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "IsConnected");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_is_connected, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

void
empd_empdd_status_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Status");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empdd_config(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_Config_Cb cb = data;
   const char *error, *error_msg;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   const char *music_directory = NULL;
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, music_directory);
        return;
     }
   if (!eldbus_message_arguments_get(msg, "s", &music_directory))
     {
        Eldbus_Error_Info error_info = {"", ""};
        ERR("Error: Getting arguments from message.");
        cb(proxy, user_data, pending, &error_info, music_directory);
        return;
     }
   cb(proxy, user_data, pending, NULL, music_directory);
   return;
}

Eldbus_Pending *
empd_empdd_config_call(Eldbus_Proxy *proxy, Empd_Empdd_Config_Cb cb, const void *data)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "Config");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_config, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

void
empd_empdd_current_song_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "CurrentSong");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_play_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Play");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_pause_call(Eldbus_Proxy *proxy, Eina_Bool mode)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Pause");
   if (!eldbus_message_arguments_append(msg, "b", mode))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_toggle_pause_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "TogglePause");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_stop_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Stop");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_next_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Next");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_previous_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Previous");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_seek_call(Eldbus_Proxy *proxy, unsigned int songid, unsigned int seconds)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Seek");
   if (!eldbus_message_arguments_append(msg, "uu", songid, seconds))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_play_id_call(Eldbus_Proxy *proxy, unsigned int songid)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "PlayId");
   if (!eldbus_message_arguments_append(msg, "u", songid))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_play_pos_call(Eldbus_Proxy *proxy, unsigned int songpos)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "PlayPos");
   if (!eldbus_message_arguments_append(msg, "u", songpos))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_queue_list_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "QueueList");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_queue_list_range_call(Eldbus_Proxy *proxy, int start, int num)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "QueueListRange");
   if (!eldbus_message_arguments_append(msg, "ii", start, num))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empdd_queue_list_cached(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_Queue_List_Cached_Cb cb = data;
   const char *error, *error_msg;
   Eina_Value *value;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, NULL);
        return;
     }
   value = eldbus_message_to_eina_value(msg);
   cb(proxy, user_data, pending, NULL, value);
   eina_value_free(value);
   return;
}

Eldbus_Pending *
empd_empdd_queue_list_cached_call(Eldbus_Proxy *proxy, Empd_Empdd_Queue_List_Cached_Cb cb, const void *data)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "QueueListCached");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_queue_list_cached, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

static void
cb_empd_empdd_queue_list_cached_range(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_Queue_List_Cached_Range_Cb cb = data;
   const char *error, *error_msg;
   Eina_Value *value;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, NULL);
        return;
     }
   value = eldbus_message_to_eina_value(msg);
   cb(proxy, user_data, pending, NULL, value);
   eina_value_free(value);
   return;
}

Eldbus_Pending *
empd_empdd_queue_list_cached_range_call(Eldbus_Proxy *proxy, Empd_Empdd_Queue_List_Cached_Range_Cb cb, const void *data, int start, int num)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "QueueListCachedRange");
   if (!eldbus_message_arguments_append(msg, "ii", start, num))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_queue_list_cached_range, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

void
empd_empdd_clear_list_call(Eldbus_Proxy *proxy)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "ClearList");
   if (!eldbus_message_arguments_append(msg, ""))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_delete_list_range_call(Eldbus_Proxy *proxy, int start, int num)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "DeleteListRange");
   if (!eldbus_message_arguments_append(msg, "ii", start, num))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_add_list_call(Eldbus_Proxy *proxy, const char *uri)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "AddList");
   if (!eldbus_message_arguments_append(msg, "s", uri))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_add_id_list_call(Eldbus_Proxy *proxy, const char *uri, unsigned int position)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "AddIdList");
   if (!eldbus_message_arguments_append(msg, "su", uri, position))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_save_playlist_call(Eldbus_Proxy *proxy, const char *uri)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "SavePlaylist");
   if (!eldbus_message_arguments_append(msg, "s", uri))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_add_to_playlist_call(Eldbus_Proxy *proxy, const char *playlist, const char *uri)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "AddToPlaylist");
   if (!eldbus_message_arguments_append(msg, "ss", playlist, uri))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_load_playlist_call(Eldbus_Proxy *proxy, const char *uri)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "LoadPlaylist");
   if (!eldbus_message_arguments_append(msg, "s", uri))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_move_list_call(Eldbus_Proxy *proxy, unsigned int start_pos, int num, unsigned int target_pos)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "MoveList");
   if (!eldbus_message_arguments_append(msg, "uiu", start_pos, num, target_pos))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_repeat_call(Eldbus_Proxy *proxy, Eina_Bool mode)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Repeat");
   if (!eldbus_message_arguments_append(msg, "b", mode))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_shuffle_call(Eldbus_Proxy *proxy, Eina_Bool mode)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Shuffle");
   if (!eldbus_message_arguments_append(msg, "b", mode))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_single_call(Eldbus_Proxy *proxy, Eina_Bool mode)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Single");
   if (!eldbus_message_arguments_append(msg, "b", mode))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

void
empd_empdd_consume_call(Eldbus_Proxy *proxy, Eina_Bool mode)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Consume");
   if (!eldbus_message_arguments_append(msg, "b", mode))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empdd_list_info(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_List_Info_Cb cb = data;
   const char *error, *error_msg;
   Eina_Value *value;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, NULL);
        return;
     }
   value = eldbus_message_to_eina_value(msg);
   cb(proxy, user_data, pending, NULL, value);
   eina_value_free(value);
   return;
}

Eldbus_Pending *
empd_empdd_list_info_call(Eldbus_Proxy *proxy, Empd_Empdd_List_Info_Cb cb, const void *data, const char *path)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "ListInfo");
   if (!eldbus_message_arguments_append(msg, "s", path))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_list_info, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

void
empd_empdd_update_call(Eldbus_Proxy *proxy, const char *path)
{
   Eldbus_Message *msg;
   EINA_SAFETY_ON_NULL_RETURN(proxy);
   msg = eldbus_proxy_method_call_new(proxy, "Update");
   if (!eldbus_message_arguments_append(msg, "s", path))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return;
     }
   eldbus_proxy_send(proxy, msg, NULL, NULL, -1);
}

static void
cb_empd_empdd_find(void *data, const Eldbus_Message *msg, Eldbus_Pending *pending)
{
   void *user_data = eldbus_pending_data_del(pending, "__user_data");
   Empd_Empdd_Find_Cb cb = data;
   const char *error, *error_msg;
   Eina_Value *value;
   Eldbus_Proxy *proxy = eldbus_pending_data_del(pending, "__proxy");
   if (eldbus_message_error_get(msg, &error, &error_msg))
     {
        Eldbus_Error_Info error_info = {error, error_msg};
        cb(proxy, user_data, pending, &error_info, NULL);
        return;
     }
   value = eldbus_message_to_eina_value(msg);
   cb(proxy, user_data, pending, NULL, value);
   eina_value_free(value);
   return;
}

Eldbus_Pending *
empd_empdd_find_call(Eldbus_Proxy *proxy, Empd_Empdd_Find_Cb cb, const void *data, unsigned int type, const char *param)
{
   Eldbus_Message *msg;
   Eldbus_Pending *p;
   EINA_SAFETY_ON_NULL_RETURN_VAL(proxy, NULL);
   msg = eldbus_proxy_method_call_new(proxy, "Find");
   if (!eldbus_message_arguments_append(msg, "us", type, param))
     {
        ERR("Error: Filling message.");
        eldbus_message_unref(msg);
        return NULL;
     }
   p = eldbus_proxy_send(proxy, msg, cb_empd_empdd_find, cb, -1);
   if (data)
     eldbus_pending_data_set(p, "__user_data", data);
   eldbus_pending_data_set(p, "__proxy", proxy);
   return p;
}

static void
empd_empdd_status_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_Status_Data *s_data = func_data;
   free(s_data);
}

static void
on_empd_empdd_status(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_Status_Data *s_data = calloc(1, sizeof(Empd_Empdd_Status_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "ibbbbuuduiiuuuuuuiiu", &s_data->volume, &s_data->repeat, &s_data->random, &s_data->single, &s_data->consume, &s_data->queue_version, &s_data->queue_length, &s_data->mixrampdb, &s_data->state, &s_data->song_pos, &s_data->songid, &s_data->song_length, &s_data->song_elapsed, &s_data->bitrate, &s_data->sample_rate, &s_data->sample_bits, &s_data->sample_channels, &s_data->next_song_pos, &s_data->next_songid, &s_data->update_id))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   ecore_event_add(EMPD_EMPDD_STATUS_EVENT, s_data, empd_empdd_status_data_free, NULL);
}

static void
empd_empdd_current_song_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_CurrentSong_Data *s_data = func_data;
   free(s_data->uri);
   free(s_data->artist);
   free(s_data->composer);
   free(s_data->title);
   free(s_data->album);
   free(s_data->name);
   free(s_data->date);
   free(s_data->disc);
   free(s_data->genre);
   free(s_data);
}

static void
on_empd_empdd_current_song(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_CurrentSong_Data *s_data = calloc(1, sizeof(Empd_Empdd_CurrentSong_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "stussssissssii", &s_data->uri, &s_data->last_modified, &s_data->duration, &s_data->artist, &s_data->composer, &s_data->title, &s_data->album, &s_data->track, &s_data->name, &s_data->date, &s_data->disc, &s_data->genre, &s_data->song_pos, &s_data->songid))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   s_data->uri = strdup(s_data->uri);
   s_data->artist = strdup(s_data->artist);
   s_data->composer = strdup(s_data->composer);
   s_data->title = strdup(s_data->title);
   s_data->album = strdup(s_data->album);
   s_data->name = strdup(s_data->name);
   s_data->date = strdup(s_data->date);
   s_data->disc = strdup(s_data->disc);
   s_data->genre = strdup(s_data->genre);
   ecore_event_add(EMPD_EMPDD_CURRENT_SONG_EVENT, s_data, empd_empdd_current_song_data_free, NULL);
}

static void
empd_empdd_queue_list_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_QueueList_Data *s_data = func_data;
   eina_value_free(s_data->value);
   free(s_data);
}

static void
on_empd_empdd_queue_list(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_QueueList_Data *s_data = calloc(1, sizeof(Empd_Empdd_QueueList_Data));
   s_data->proxy = proxy;
   s_data->value = eldbus_message_to_eina_value(msg);
   ecore_event_add(EMPD_EMPDD_QUEUE_LIST_EVENT, s_data, empd_empdd_queue_list_data_free, NULL);
}

static void
empd_empdd_login_failed_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_LoginFailed_Data *s_data = func_data;
   free(s_data->host);
   free(s_data);
}

static void
on_empd_empdd_login_failed(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_LoginFailed_Data *s_data = calloc(1, sizeof(Empd_Empdd_LoginFailed_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "su", &s_data->host, &s_data->port))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   s_data->host = strdup(s_data->host);
   ecore_event_add(EMPD_EMPDD_LOGIN_FAILED_EVENT, s_data, empd_empdd_login_failed_data_free, NULL);
}

static void
empd_empdd_permission_failed_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_PermissionFailed_Data *s_data = func_data;
   free(s_data->host);
   free(s_data);
}

static void
on_empd_empdd_permission_failed(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_PermissionFailed_Data *s_data = calloc(1, sizeof(Empd_Empdd_PermissionFailed_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "suu", &s_data->host, &s_data->port, &s_data->code))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   s_data->host = strdup(s_data->host);
   ecore_event_add(EMPD_EMPDD_PERMISSION_FAILED_EVENT, s_data, empd_empdd_permission_failed_data_free, NULL);
}

static void
empd_empdd_connected_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_Connected_Data *s_data = func_data;
   free(s_data->host);
   free(s_data);
}

static void
on_empd_empdd_connected(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_Connected_Data *s_data = calloc(1, sizeof(Empd_Empdd_Connected_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, "su", &s_data->host, &s_data->port))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   s_data->host = strdup(s_data->host);
   ecore_event_add(EMPD_EMPDD_CONNECTED_EVENT, s_data, empd_empdd_connected_data_free, NULL);
}

static void
empd_empdd_disconnected_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_Disconnected_Data *s_data = func_data;
   free(s_data);
}

static void
on_empd_empdd_disconnected(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_Disconnected_Data *s_data = calloc(1, sizeof(Empd_Empdd_Disconnected_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, ""))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   ecore_event_add(EMPD_EMPDD_DISCONNECTED_EVENT, s_data, empd_empdd_disconnected_data_free, NULL);
}

static void
empd_empdd_database_update_begin_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_DatabaseUpdateBegin_Data *s_data = func_data;
   free(s_data);
}

static void
on_empd_empdd_database_update_begin(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_DatabaseUpdateBegin_Data *s_data = calloc(1, sizeof(Empd_Empdd_DatabaseUpdateBegin_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, ""))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   ecore_event_add(EMPD_EMPDD_DATABASE_UPDATE_BEGIN_EVENT, s_data, empd_empdd_database_update_begin_data_free, NULL);
}

static void
empd_empdd_database_update_end_data_free(void *user_data EINA_UNUSED, void *func_data)
{
   Empd_Empdd_DatabaseUpdateEnd_Data *s_data = func_data;
   free(s_data);
}

static void
on_empd_empdd_database_update_end(void *data, const Eldbus_Message *msg)
{
   Eldbus_Proxy *proxy = data;
   Empd_Empdd_DatabaseUpdateEnd_Data *s_data = calloc(1, sizeof(Empd_Empdd_DatabaseUpdateEnd_Data));
   s_data->proxy = proxy;
   if (!eldbus_message_arguments_get(msg, ""))
     {
        ERR("Error: Getting arguments from message.");
        free(s_data);
        return;
     }
   ecore_event_add(EMPD_EMPDD_DATABASE_UPDATE_END_EVENT, s_data, empd_empdd_database_update_end_data_free, NULL);
}

void
empd_empdd_log_domain_set(int id)
{
   _log_main = id;
}

void
empd_empdd_proxy_unref(Eldbus_Proxy *proxy)
{
   Eldbus_Object *obj = eldbus_proxy_object_get(proxy);
   eldbus_proxy_unref(proxy);
   eldbus_object_unref(obj);
}

Eldbus_Proxy *
empd_empdd_proxy_get(Eldbus_Connection *conn, const char *bus, const char *path)
{
   Eldbus_Object *obj;
   Eldbus_Proxy *proxy;
   EINA_SAFETY_ON_NULL_RETURN_VAL(conn, NULL);
   EINA_SAFETY_ON_NULL_RETURN_VAL(bus, NULL);
   if (!path) path = "/";
   obj = eldbus_object_get(conn, bus, path);
   proxy = eldbus_proxy_get(obj, "org.empd.empdd");
   eldbus_proxy_signal_handler_add(proxy, "Status", on_empd_empdd_status, proxy);
   if (!EMPD_EMPDD_STATUS_EVENT)
     EMPD_EMPDD_STATUS_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "CurrentSong", on_empd_empdd_current_song, proxy);
   if (!EMPD_EMPDD_CURRENT_SONG_EVENT)
     EMPD_EMPDD_CURRENT_SONG_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "QueueList", on_empd_empdd_queue_list, proxy);
   if (!EMPD_EMPDD_QUEUE_LIST_EVENT)
     EMPD_EMPDD_QUEUE_LIST_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "LoginFailed", on_empd_empdd_login_failed, proxy);
   if (!EMPD_EMPDD_LOGIN_FAILED_EVENT)
     EMPD_EMPDD_LOGIN_FAILED_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "PermissionFailed", on_empd_empdd_permission_failed, proxy);
   if (!EMPD_EMPDD_PERMISSION_FAILED_EVENT)
     EMPD_EMPDD_PERMISSION_FAILED_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "Connected", on_empd_empdd_connected, proxy);
   if (!EMPD_EMPDD_CONNECTED_EVENT)
     EMPD_EMPDD_CONNECTED_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "Disconnected", on_empd_empdd_disconnected, proxy);
   if (!EMPD_EMPDD_DISCONNECTED_EVENT)
     EMPD_EMPDD_DISCONNECTED_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "DatabaseUpdateBegin", on_empd_empdd_database_update_begin, proxy);
   if (!EMPD_EMPDD_DATABASE_UPDATE_BEGIN_EVENT)
     EMPD_EMPDD_DATABASE_UPDATE_BEGIN_EVENT = ecore_event_type_new();
   eldbus_proxy_signal_handler_add(proxy, "DatabaseUpdateEnd", on_empd_empdd_database_update_end, proxy);
   if (!EMPD_EMPDD_DATABASE_UPDATE_END_EVENT)
     EMPD_EMPDD_DATABASE_UPDATE_END_EVENT = ecore_event_type_new();
   return proxy;
}
