#ifndef ELDBUS_EMPD_EMPDD_H
#define ELDBUS_EMPD_EMPDD_H

#include <Eina.h>
#include <Ecore.h>
#include <Eldbus.h>
#include "eldbus_utils.h"

Eldbus_Proxy *empd_empdd_proxy_get(Eldbus_Connection *conn, const char *bus, const char *path);
void empd_empdd_proxy_unref(Eldbus_Proxy *proxy);
void empd_empdd_log_domain_set(int id);
void empd_empdd_connect_call(Eldbus_Proxy *proxy, const char *host, int port, const char *password);
typedef void (*Empd_Empdd_Is_Connected_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, Eina_Bool isconnected, const char *host, unsigned int port);
Eldbus_Pending *empd_empdd_is_connected_call(Eldbus_Proxy *proxy, Empd_Empdd_Is_Connected_Cb cb, const void *data);
void empd_empdd_status_call(Eldbus_Proxy *proxy);
typedef void (*Empd_Empdd_Config_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, const char *music_directory);
Eldbus_Pending *empd_empdd_config_call(Eldbus_Proxy *proxy, Empd_Empdd_Config_Cb cb, const void *data);
void empd_empdd_current_song_call(Eldbus_Proxy *proxy);
void empd_empdd_play_call(Eldbus_Proxy *proxy);
void empd_empdd_pause_call(Eldbus_Proxy *proxy, Eina_Bool mode);
void empd_empdd_toggle_pause_call(Eldbus_Proxy *proxy);
void empd_empdd_stop_call(Eldbus_Proxy *proxy);
void empd_empdd_next_call(Eldbus_Proxy *proxy);
void empd_empdd_previous_call(Eldbus_Proxy *proxy);
void empd_empdd_seek_call(Eldbus_Proxy *proxy, unsigned int songid, unsigned int seconds);
void empd_empdd_play_id_call(Eldbus_Proxy *proxy, unsigned int songid);
void empd_empdd_play_pos_call(Eldbus_Proxy *proxy, unsigned int songpos);
void empd_empdd_queue_list_call(Eldbus_Proxy *proxy);
void empd_empdd_queue_list_range_call(Eldbus_Proxy *proxy, int start, int num);
typedef void (*Empd_Empdd_Queue_List_Cached_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, Eina_Value *args);
Eldbus_Pending *empd_empdd_queue_list_cached_call(Eldbus_Proxy *proxy, Empd_Empdd_Queue_List_Cached_Cb cb, const void *data);
typedef void (*Empd_Empdd_Queue_List_Cached_Range_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, Eina_Value *args);
Eldbus_Pending *empd_empdd_queue_list_cached_range_call(Eldbus_Proxy *proxy, Empd_Empdd_Queue_List_Cached_Range_Cb cb, const void *data, int start, int num);
void empd_empdd_clear_list_call(Eldbus_Proxy *proxy);
void empd_empdd_delete_list_range_call(Eldbus_Proxy *proxy, int start, int num);
void empd_empdd_add_list_call(Eldbus_Proxy *proxy, const char *uri);
void empd_empdd_add_id_list_call(Eldbus_Proxy *proxy, const char *uri, unsigned int position);
void empd_empdd_save_playlist_call(Eldbus_Proxy *proxy, const char *uri);
void empd_empdd_add_to_playlist_call(Eldbus_Proxy *proxy, const char *playlist, const char *uri);
void empd_empdd_load_playlist_call(Eldbus_Proxy *proxy, const char *uri);
void empd_empdd_move_list_call(Eldbus_Proxy *proxy, unsigned int start_pos, int num, unsigned int target_pos);
void empd_empdd_repeat_call(Eldbus_Proxy *proxy, Eina_Bool mode);
void empd_empdd_shuffle_call(Eldbus_Proxy *proxy, Eina_Bool mode);
void empd_empdd_single_call(Eldbus_Proxy *proxy, Eina_Bool mode);
void empd_empdd_consume_call(Eldbus_Proxy *proxy, Eina_Bool mode);
typedef void (*Empd_Empdd_List_Info_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, Eina_Value *args);
Eldbus_Pending *empd_empdd_list_info_call(Eldbus_Proxy *proxy, Empd_Empdd_List_Info_Cb cb, const void *data, const char *path);
void empd_empdd_update_call(Eldbus_Proxy *proxy, const char *path);
typedef void (*Empd_Empdd_Find_Cb)(Eldbus_Proxy *proxy, void *data, Eldbus_Pending *pending, Eldbus_Error_Info *error, Eina_Value *args);
Eldbus_Pending *empd_empdd_find_call(Eldbus_Proxy *proxy, Empd_Empdd_Find_Cb cb, const void *data, unsigned int type, const char *param);
extern int EMPD_EMPDD_STATUS_EVENT;
typedef struct _Empd_Empdd_Status_Data
{
   Eldbus_Proxy *proxy;
   int volume;
   Eina_Bool repeat;
   Eina_Bool random;
   Eina_Bool single;
   Eina_Bool consume;
   unsigned int queue_version;
   unsigned int queue_length;
   double mixrampdb;
   unsigned int state;
   int song_pos;
   int songid;
   unsigned int song_length;
   unsigned int song_elapsed;
   unsigned int bitrate;
   unsigned int sample_rate;
   unsigned int sample_bits;
   unsigned int sample_channels;
   int next_song_pos;
   int next_songid;
   unsigned int update_id;
} Empd_Empdd_Status_Data;
extern int EMPD_EMPDD_CURRENT_SONG_EVENT;
typedef struct _Empd_Empdd_CurrentSong_Data
{
   Eldbus_Proxy *proxy;
   char *uri;
   uint64_t last_modified;
   unsigned int duration;
   char *artist;
   char *composer;
   char *title;
   char *album;
   int track;
   char *name;
   char *date;
   char *disc;
   char *genre;
   int song_pos;
   int songid;
} Empd_Empdd_CurrentSong_Data;
extern int EMPD_EMPDD_QUEUE_LIST_EVENT;
typedef struct _Empd_Empdd_QueueList_Data
{
   Eldbus_Proxy *proxy;
   Eina_Value *value;
} Empd_Empdd_QueueList_Data;
extern int EMPD_EMPDD_LOGIN_FAILED_EVENT;
typedef struct _Empd_Empdd_LoginFailed_Data
{
   Eldbus_Proxy *proxy;
   char *host;
   unsigned int port;
} Empd_Empdd_LoginFailed_Data;
extern int EMPD_EMPDD_PERMISSION_FAILED_EVENT;
typedef struct _Empd_Empdd_PermissionFailed_Data
{
   Eldbus_Proxy *proxy;
   char *host;
   unsigned int port;
   unsigned int code;
} Empd_Empdd_PermissionFailed_Data;
extern int EMPD_EMPDD_CONNECTED_EVENT;
typedef struct _Empd_Empdd_Connected_Data
{
   Eldbus_Proxy *proxy;
   char *host;
   unsigned int port;
} Empd_Empdd_Connected_Data;
extern int EMPD_EMPDD_DISCONNECTED_EVENT;
typedef struct _Empd_Empdd_Disconnected_Data
{
   Eldbus_Proxy *proxy;
} Empd_Empdd_Disconnected_Data;
extern int EMPD_EMPDD_DATABASE_UPDATE_BEGIN_EVENT;
typedef struct _Empd_Empdd_DatabaseUpdateBegin_Data
{
   Eldbus_Proxy *proxy;
} Empd_Empdd_DatabaseUpdateBegin_Data;
extern int EMPD_EMPDD_DATABASE_UPDATE_END_EVENT;
typedef struct _Empd_Empdd_DatabaseUpdateEnd_Data
{
   Eldbus_Proxy *proxy;
} Empd_Empdd_DatabaseUpdateEnd_Data;

#endif