#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include "Lyricwiki_Common.h"
#include <string.h>
#include <inttypes.h>
#include <errno.h>

Eina_Bool
azy_str_to_bool_(const char *d, Eina_Bool *ret)
{
  *ret = EINA_FALSE;
  if (d && (*d == '1')) *ret = EINA_TRUE;
  return EINA_TRUE;
}

Eina_Bool
azy_str_to_str_(const char *d, const char **ret)
{
  *ret = NULL;
  if (!d) return EINA_TRUE;
  *ret = eina_stringshare_add(d);
  return EINA_TRUE;
}

Eina_Bool
azy_str_to_int_(const char *d, int *ret)
{
  errno = 0;
  *ret = 0;
  if (!d) return EINA_TRUE;
  *ret = strtol(d, NULL, 10);
  if (errno)
    {
#ifdef ERR
      fprintf(stderr, "Error converting %s to int: '%s'", d, strerror(errno));
#endif
      return EINA_FALSE;
    }
  return EINA_TRUE;
}

Eina_Bool
azy_str_to_double_(const char *d, double *ret)
{
  errno = 0;
  *ret = 0.0;
  if (!d) return EINA_TRUE;
  *ret = strtod(d, NULL);
  if (errno)
    {
#ifdef ERR
      fprintf(stderr, "Error converting %s to double: '%s'", d, strerror(errno));
#endif
      return EINA_FALSE;
    }
  return EINA_TRUE;
}
void Lyricwiki_Lyric_free(Lyricwiki_Lyric * val)
{
  if (!val)
    return;

  eina_stringshare_del(val->artist);
  eina_stringshare_del(val->song);
  eina_stringshare_del(val->lyrics);
  eina_stringshare_del(val->url);
  free(val);
}

Lyricwiki_Lyric * Lyricwiki_Lyric_copy(Lyricwiki_Lyric * orig)
{
  Lyricwiki_Lyric * copy;

  if (!orig)
    return NULL;

  copy = Lyricwiki_Lyric_new();
  copy->artist = eina_stringshare_ref(orig->artist);
  copy->song = eina_stringshare_ref(orig->song);
  copy->lyrics = eina_stringshare_ref(orig->lyrics);
  copy->url = eina_stringshare_ref(orig->url);

  return copy;
}

Eina_Bool Lyricwiki_Lyric_eq(Lyricwiki_Lyric * a, Lyricwiki_Lyric * b)
{
  if (a == b)
    return EINA_TRUE;
  if ((!a) || (!b))
    return EINA_FALSE;
  if (a->artist != b->artist)
    return EINA_FALSE;
  if (a->song != b->song)
    return EINA_FALSE;
  if (a->lyrics != b->lyrics)
    return EINA_FALSE;
  if (a->url != b->url)
    return EINA_FALSE;

  return EINA_TRUE;
}

void Lyricwiki_Lyric_print(const char *pre, int indent, const Lyricwiki_Lyric *a)
{
  int i;
  if (!a)
    return;
  if (!pre) pre = "\t";
  for (i = 0; i < indent; i++)
    printf("%s", pre);
  printf("artist: %s\n", a->artist);
  for (i = 0; i < indent; i++)
    printf("%s", pre);
  printf("song: %s\n", a->song);
  for (i = 0; i < indent; i++)
    printf("%s", pre);
  printf("lyrics: %s\n", a->lyrics);
  for (i = 0; i < indent; i++)
    printf("%s", pre);
  printf("url: %s\n", a->url);
}

Eina_Bool Lyricwiki_Lyric_isnull(Lyricwiki_Lyric * a)
{
  if (!a)
    return EINA_TRUE;
  if (a->artist != NULL)
    return EINA_FALSE;
  if (a->song != NULL)
    return EINA_FALSE;
  if (a->lyrics != NULL)
    return EINA_FALSE;
  if (a->url != NULL)
    return EINA_FALSE;

  return EINA_TRUE;
}

